import os

from IPython import embed

import numpy

from astropy.io import fits

from mangadap.config.manga import MaNGAConfig
from mangadap.datacube import MaNGADataCube
from mangadap.util.covariance import Covariance
from mangadap.util.constants import DAPConstants
from mangadap.tests.util import requires_remote, remote_data_file

#-----------------------------------------------------------------------------

def test_samples():

    rng = numpy.random.default_rng(seed=8001)
    
    # Build a bogus covariance matrix
    m = numpy.zeros(10, dtype=float)
    c = numpy.diag(numpy.full(10-2, 0.2, dtype=float), k=-2) \
            + numpy.diag(numpy.full(10-1, 0.5, dtype=float), k=-1) \
            + numpy.diag(numpy.full(10, 1.0, dtype=float), k=0) \
            + numpy.diag(numpy.full(10-1, 0.5, dtype=float), k=1) \
            + numpy.diag(numpy.full(10-2, 0.2, dtype=float), k=2)

    # Draw samples
    s = rng.multivariate_normal(m, c, size=100000)

    # Instantiate
    covar = Covariance.from_samples(s.T, cov_tol=0.1)

    # Check the values are very nearly the same as the input
    assert numpy.all(numpy.absolute(c - covar.toarray()) < 0.02), 'Covariances are too different'

    # Check that `find` returns the same indices
    # NOTE: For some reason the ordering of numpy.where and
    # scipy.sparse.find are different.  So I run lexsort before checking
    # if the arrays are the same.

    coo = numpy.array([ [i,j] for i,j in zip(*numpy.where(c > 0))])
    srt = numpy.lexsort((coo[:,0], coo[:,1]))
    coo = coo[srt,:]

    _coo = numpy.array([ [i,j] for i,j,k in zip(*covar.find())])
    srt = numpy.lexsort((_coo[:,0], _coo[:,1]))
    _coo = _coo[srt,:]

    assert numpy.array_equal(coo, _coo), 'Did not find the same indices'


def test_mult():

    # Build a bogus covariance matrix
    c = numpy.diag(numpy.full(10-2, 0.2, dtype=float), k=-2) \
            + numpy.diag(numpy.full(10-1, 0.5, dtype=float), k=-1) \
            + numpy.diag(numpy.full(10, 1.0, dtype=float), k=0) \
            + numpy.diag(numpy.full(10-1, 0.5, dtype=float), k=1) \
            + numpy.diag(numpy.full(10-2, 0.2, dtype=float), k=2)

    x = numpy.ones(10, dtype=float)

    # Uncorrelated
    t = numpy.zeros((3,10), dtype=float)
    t[0,0] = 1.0
    t[1,3] = 1.0
    t[2,6] = 1.0

    y = numpy.dot(t, x)

    covar = Covariance.from_matrix_multiplication(t, c)
    assert numpy.array_equal(covar.toarray(), numpy.identity(3)), 'Result should be uncorrelated.'

    # Correlated by 0.2
    t = numpy.zeros((3,10), dtype=float)
    t[0,0] = 1.0
    t[1,2] = 1.0
    t[2,4] = 1.0
    _c = numpy.diag(numpy.full(3-1, 0.2, dtype=float), k=-1) \
            + numpy.diag(numpy.full(3, 1.0, dtype=float), k=0) \
            + numpy.diag(numpy.full(3-1, 0.2, dtype=float), k=1)
    y = numpy.dot(t, x)
    covar = Covariance.from_matrix_multiplication(t, c)
    assert numpy.array_equal(covar.toarray(), _c), 'Result should have off-diagonals = 0.2'

    # Correlated by 0.5 and 0.2
    t = numpy.zeros((3,10), dtype=float)
    t[0,0] = 1.0
    t[1,1] = 1.0
    t[2,2] = 1.0
    _c = numpy.diag(numpy.full(3-2, 0.2, dtype=float), k=-2) \
            + numpy.diag(numpy.full(3-1, 0.5, dtype=float), k=-1) \
            + numpy.diag(numpy.full(3, 1.0, dtype=float), k=0) \
            + numpy.diag(numpy.full(3-1, 0.5, dtype=float), k=1) \
            + numpy.diag(numpy.full(3-2, 0.2, dtype=float), k=2)
    y = numpy.dot(t, x)
    covar = Covariance.from_matrix_multiplication(t, c)
    assert numpy.array_equal(covar.toarray(), _c), 'Result should have off-diagonals = 0.5,0.2'


def test_var():
    var = numpy.ones(3, dtype=float)
    covar = Covariance.from_variance(var)
    assert numpy.array_equal(covar.toarray(), numpy.identity(3)), \
            'Result should be an identity matrix'


def test_array():
    # Construct the Covariance matrix from a pre-calculated array
    c = numpy.diag(numpy.full(10-2, 0.2, dtype=float), k=-2) \
            + numpy.diag(numpy.full(10-1, 0.5, dtype=float), k=-1) \
            + numpy.diag(numpy.full(10, 1.0, dtype=float), k=0) \
            + numpy.diag(numpy.full(10-1, 0.5, dtype=float), k=1) \
            + numpy.diag(numpy.full(10-2, 0.2, dtype=float), k=2)
    covar = Covariance.from_array(c)
    # Should be the same as the identity matrix.
    assert numpy.array_equal(covar.toarray(), c), 'Arrays should be identical'


def test_io():

    rng = numpy.random.default_rng(seed=8001)
    
    # Clean up in case of a failure
    ofile = 'test_covar_io.fits'
    if os.path.isfile(ofile):
        os.remove(ofile)

    # Build a bogus covariance matrix
    m = numpy.zeros(10, dtype=float)
    c = numpy.diag(numpy.full(10-2, 0.2, dtype=float), k=-2) \
            + numpy.diag(numpy.full(10-1, 0.5, dtype=float), k=-1) \
            + numpy.diag(numpy.full(10, 1.0, dtype=float), k=0) \
            + numpy.diag(numpy.full(10-1, 0.5, dtype=float), k=1) \
            + numpy.diag(numpy.full(10-2, 0.2, dtype=float), k=2)

    # Draw samples
    s = rng.multivariate_normal(m, c, size=100000)

    # Instantiate
    covar = Covariance.from_samples(s.T, cov_tol=0.1)
    # Write
    covar.write(ofile)
    # Read
    _covar = Covariance.from_fits(ofile)
    # Should be the same
    assert numpy.allclose(covar.toarray(), _covar.toarray()), 'Bad I/O'
    # Clean-up 
    os.remove(ofile)


@requires_remote
def test_read_drp():
    cfg = MaNGAConfig(7815, 3702)
    drpfile = remote_data_file(cfg.file_name)
    assert os.path.isfile(drpfile), 'Did not find file'

    with fits.open(drpfile) as hdu:
        covar = Covariance.from_fits(hdu, ivar_ext=None, covar_ext='GCORREL', impose_triu=True,
                                     correlation=True)
        var = numpy.ma.power(hdu['IVAR'].data[hdu['GCORREL'].header['BBINDEX']].T.ravel(),
                             -1).filled(0.0)

    covar = covar.apply_new_variance(var)
    covar.revert_correlation()

    assert numpy.array_equal(var, numpy.diag(covar.toarray())), 'New variance not applied'


@requires_remote
def test_rectification_recovery():
    cube = MaNGADataCube.from_plateifu(7815, 3702, directory_path=remote_data_file(),
                                       covar_ext='GCORREL')
    cube.load_rss()

    hdu = fits.open(str(cube.file_path))
    channel = hdu['GCORREL'].header['BBINDEX']

    gcorrel = numpy.zeros(eval(hdu['GCORREL'].header['COVSHAPE']), dtype=float)
    i = numpy.ravel_multi_index((hdu['GCORREL'].data['INDXI_C1'],hdu['GCORREL'].data['INDXI_C2']),
                                cube.spatial_shape)

    j = numpy.ravel_multi_index((hdu['GCORREL'].data['INDXJ_C1'],hdu['GCORREL'].data['INDXJ_C2']),
                                cube.spatial_shape)
    gcorrel[i,j] = hdu['GCORREL'].data['RHOIJ']
    gcorrel[j,i] = hdu['GCORREL'].data['RHOIJ']

    assert numpy.allclose(cube.covar.toarray(), gcorrel), 'Bad covariance read'

    flux, C = cube.rss.rectify_wavelength_plane(channel, return_covar=True)
    assert numpy.allclose(cube.flux[...,channel], flux), 'Bad flux rectification'

    ivar = numpy.ma.power(C.variance().reshape(cube.spatial_shape), -1).filled(0.0)
    assert numpy.allclose(cube.ivar[...,channel], ivar), 'Bad inverse variance rectification'

    C.to_correlation()
    assert numpy.allclose(C.toarray(), gcorrel), 'Bad covariance calculation'

    sres = numpy.ma.divide(cube.rss.wave[channel],
                           cube.rss.instrumental_dispersion_plane(channel).ravel()) \
                / DAPConstants.sig2fwhm

    # WARNING: The computations done by the DRP and DAP are different
    # in detail, but (at least for this test cube) the results are
    # virtually identical except for notable outliers.
    assert numpy.ma.median(cube.sres[...,channel].ravel() - sres) < 0.1, \
            'Bad spectral resolution rectification'

#    from matplotlib import pyplot
#    zoom = 12
#    xs = int(C.shape[0]/2 - C.shape[0]/2/zoom)
#    xe = xs + int(C.shape[0]/zoom) + 1
#    ys = int(C.shape[1]/2 - C.shape[1]/2/zoom)
#    ye = ys + int(C.shape[1]/zoom) + 1
#
#    fig = pyplot.figure(figsize=(pyplot.figaspect(1)))
#    ax = fig.add_axes([0.1, 0.3, 0.35, 0.35])
#    ax.imshow(C.toarray()[xs:xe,ys:ye], origin='lower', interpolation='nearest',
#              cmap='inferno', aspect='auto')
#    ax.set_title('DAP')
#    ax = fig.add_axes([0.55, 0.3, 0.35, 0.35])
#    cax = fig.add_axes([0.91, 0.3, 0.01, 0.35])
#    ax.set_title('DRP')
#    cs = ax.imshow(cube.covar.toarray()[xs:xe,ys:ye], origin='lower',
#                   interpolation='nearest', aspect='auto', cmap='inferno')
#    pyplot.colorbar(cs, cax=cax)
#    pyplot.show()





import os
import time

from IPython import embed

import numpy

from astropy.io import fits

from matplotlib.backends.backend_pdf import PdfPages
from matplotlib import pyplot, rc, colors, ticker

from mangadap.util.drpfits import DRPFitsBitMask
from mangadap.dapfits import DAPCubeBitMask
from mangadap.proc.stellarcontinuummodel import StellarContinuumModel

#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------

def main():
    select_file = 'repr/representative_spectra_selection_v2.fits'
    spec_file = 'repr/benchmark_spectra_v2.fits'
    flag_file = 'repr/representative_spectra_flags_v2.db'

#    mod_file = None
#    plot_file = 'repr/representative_spectra_v2.pdf'

#    mod_file = 'repr/benchmark_spectra_v2_model.fits'
#    plot_file = 'repr/representative_spectra_v2_model.pdf'

#    mod_file = 'repr/run5/run5_MILESHC_4_8_MASTARSSP_1_14_ELBMPL9_ELPMPL11.fits.gz'
#    plot_file = 'repr/representative_spectra_v2_model_run5_ssp_14_sig1.4.pdf'
#    mod_file = 'repr/run5/run5_MILESHC_4_8_MASTARSSP_1_14_ELBMPL9_ELPMPL11A.fits.gz'
#    plot_file = 'repr/representative_spectra_v2_model_run5_ssp_14_sig3.0.pdf'
#    mod_file = 'repr/run5/run5_MILESHC_4_8_MASTARSSP_1_14_ELBMPL9_ELPMPL11B.fits.gz'
#    plot_file = 'repr/representative_spectra_v2_model_run5_ssp_14_sigmix.pdf'
    mod_file = 'repr/run7/run7_MILESHC_4_8_MASTARSSP_1_14_ELBMPL9_ELPMPL11UPA.fits.gz'
    plot_file = 'repr/representative_spectra_v2_model_run7_ssp_14_sigmixup.pdf'

    select_hdu = fits.open(select_file)
    spec_hdu = fits.open(spec_file)
    mod_hdu = None if mod_file is None else fits.open(mod_file)

    nspec = spec_hdu['FLUX'].shape[0]

    restwave_limits = numpy.array([3575, 10300])
    d4000_limits = numpy.array([3700, 4300])
    mg_limits = numpy.array([5100, 5285])
    halpha_limits = numpy.array([6460, 6676])
    ca_limits = numpy.array([8450, 8750])

    drp_bm = DRPFitsBitMask()
    cube_bm = DAPCubeBitMask()

    wave = spec_hdu['WAVE'].data
    flux = numpy.ma.MaskedArray(spec_hdu['FLUX'].data,
                                mask=drp_bm.flagged(spec_hdu['MASK'].data,
                                                flag=['DONOTUSE', 'FORESTAR']))

    if mod_hdu is None:
        model = numpy.ma.MaskedArray(spec_hdu['MODEL'].data,
                                     mask=cube_bm.flagged(spec_hdu['MODEL_MASK'].data,
                                                      flag='NOMODEL'))
        cont = numpy.ma.MaskedArray(spec_hdu['MODEL'].data - spec_hdu['EMLINE'].data,
                                    mask=cube_bm.flagged(spec_hdu['MODEL_MASK'].data,
                                                      flag='NOMODEL'))
        stellar = numpy.ma.MaskedArray(spec_hdu['STELLAR'].data,
                                       mask=cube_bm.flagged(spec_hdu['STELLAR_MASK'].data,
                                                      flag='NOMODEL'))
    else:
        model = numpy.ma.MaskedArray(mod_hdu['MODEL'].data, mask=mod_hdu['MODEL_MASK'].data > 0)
        model = StellarContinuumModel.reset_continuum_mask_window(model)
        cont = numpy.ma.MaskedArray(mod_hdu['MODEL'].data - mod_hdu['EMLINE'].data,
                                    mask=mod_hdu['MODEL_MASK'].data > 0)
        cont = StellarContinuumModel.reset_continuum_mask_window(cont)
        
        stellar = numpy.ma.MaskedArray(mod_hdu['STELLAR'].data,
                                       mask=mod_hdu['STELLAR_MASK'].data > 0)
        stellar = StellarContinuumModel.reset_continuum_mask_window(stellar)

    if not os.path.isfile(flag_file):
        include_flags = numpy.ones(nspec, dtype=int)
        numpy.savetxt(flag_file, numpy.array([select_hdu['PAR'].data['PLT'].astype(int),
                                              select_hdu['PAR'].data['IFU'].astype(int),
                                              select_hdu['PAR'].data['BIN'].astype(int),
                                              numpy.ones(nspec, dtype=int)]).T,
                      fmt=['%5d', '%5d', '%5d', '%3d'],
                      header='{0:>3} {1:>5} {2:>5} {3:>3}'.format('PLT', 'IFU', 'BIN', 'SEL'))

    rc('font', size=8)

    with PdfPages(plot_file) as pdf:

#        j = 0
        for i in range(nspec):
            print('{0}/{1}'.format((i+1),nspec), end='\r')
            lambda_limits = restwave_limits*(1+select_hdu['PAR'].data['Z'][i])
            indx = numpy.logical_not(numpy.ma.getmaskarray(flux)[i,:]) \
                        & (wave > lambda_limits[0]) & (wave < lambda_limits[1])
            if numpy.sum(indx) == 0:
                continue
            if numpy.sum(numpy.logical_not(numpy.ma.getmaskarray(model)[i,:])) == 0:
                continue

            mod_lim = [numpy.ma.amin(model[i,indx]), numpy.ma.amax(model[i,indx])]
            Df = (mod_lim[1]-mod_lim[0])*1.5
            flux_limits = numpy.mean(mod_lim) + numpy.array([-Df/2, Df/2])

            if not numpy.all(numpy.isfinite(flux_limits)):
                embed()
                exit()

            fig = pyplot.figure()

            # Full spectrum
            ax = fig.add_axes([0.06, 0.45, 0.92, 0.5])
            ax.minorticks_on()
            ax.tick_params(which='major', length=10, direction='in', top=True, right=True)
            ax.tick_params(which='minor', length=5, direction='in', top=True, right=True)
            ax.set_xlim(lambda_limits)
            ax.set_ylim(flux_limits)
            ax.grid(True, which='major', color='0.8', zorder=0, linestyle='-')
            ax.step(wave, flux[i,:], zorder=2, color='k', lw=0.5, where='mid')
            ax.plot(wave, stellar[i,:], zorder=3, color='C1', lw=1.0)
            ax.plot(wave, model[i,:], zorder=4, color='C3', lw=1.0)
            ax.plot(wave, cont[i,:], zorder=5, color='C9', lw=1.0)

            ax.text(-0.05, 0.5, r'$F_\lambda$',
                    ha='center', va='center', transform=ax.transAxes, rotation='vertical')
            ax.text(0.5, -0.1, r'$\lambda_{\rm obs}\ [{\rm \AA}]$',
                    ha='center', va='center', transform=ax.transAxes)

            ax.text(0.01, 1.03,
                    '{0}: {1}-{2}-{3}; '.format(i, select_hdu['PAR'].data['PLT'][i],
                                                select_hdu['PAR'].data['IFU'][i],
                                                select_hdu['PAR'].data['BIN'][i])
                    + 'z={0:.4f}; '.format(select_hdu['PAR'].data['Z'][i])
                    + 'S/N={0:.1f}; '.format(select_hdu['PAR'].data['SNR'][i])
                    + r'$\sigma_{\ast,{\rm obs}}$='
                    + r'{0:.1f}; D4000={1:.1f}; H$\alpha$ EW={2:.1f}'.format(
                        select_hdu['PAR'].data['SIGMA'][i], select_hdu['PAR'].data['D4000'][i],
                        select_hdu['PAR'].data['HAEW'][i]),
                    horizontalalignment='left', verticalalignment='center', transform=ax.transAxes)


            lambda_limits = d4000_limits*(1+select_hdu['PAR'].data['Z'][i])
            indx = numpy.logical_not(numpy.ma.getmaskarray(flux)[i,:]) \
                        & (wave > lambda_limits[0]) & (wave < lambda_limits[1])
            mod_lim = [0,1] if numpy.sum(indx) == 0 \
                        else [numpy.ma.amin(model[i,indx]), numpy.ma.amax(model[i,indx])]
            Df = (mod_lim[1]-mod_lim[0])*1.5
            flux_limits = numpy.mean(mod_lim) + numpy.array([-Df/2, Df/2])

            ax = fig.add_axes([0.06, 0.08, 0.19, 0.27])
            ax.minorticks_on()
            ax.tick_params(which='major', length=10, direction='in', top=True, right=True)
            ax.tick_params(which='minor', length=5, direction='in', top=True, right=True)
            ax.set_xlim(lambda_limits)
            ax.set_ylim(flux_limits)
            ax.grid(True, which='major', color='0.8', zorder=0, linestyle='-')
            ax.plot(wave, flux[i,:], zorder=2, color='k', lw=0.5)
            ax.plot(wave, stellar[i,:], zorder=3, color='C1', lw=1.0)
            ax.plot(wave, model[i,:], zorder=4, color='C3', lw=1.0)
            ax.plot(wave, cont[i,:], zorder=5, color='C9', lw=1.0)

            ax.text(0.5, -0.2, r'$\lambda_{\rm obs}\ [{\rm \AA}]$',
                    ha='center', va='center', transform=ax.transAxes)




            lambda_limits = mg_limits*(1+select_hdu['PAR'].data['Z'][i])
            indx = numpy.logical_not(numpy.ma.getmaskarray(flux)[i,:]) \
                        & (wave > lambda_limits[0]) & (wave < lambda_limits[1])
            mod_lim = [0,1] if numpy.sum(indx) == 0 \
                        else [numpy.ma.amin(model[i,indx]), numpy.ma.amax(model[i,indx])]
            Df = (mod_lim[1]-mod_lim[0])*1.5
            flux_limits = numpy.mean(mod_lim) + numpy.array([-Df/2, Df/2])

            ax = fig.add_axes([0.30, 0.08, 0.19, 0.27])
            ax.minorticks_on()
            ax.tick_params(which='major', length=10, direction='in', top=True, right=True)
            ax.tick_params(which='minor', length=5, direction='in', top=True, right=True)
            ax.set_xlim(lambda_limits)
            ax.set_ylim(flux_limits)
            ax.grid(True, which='major', color='0.8', zorder=0, linestyle='-')
            ax.plot(wave, flux[i,:], zorder=2, color='k', lw=0.5)
            ax.plot(wave, stellar[i,:], zorder=3, color='C1', lw=1.0)
            ax.plot(wave, model[i,:], zorder=4, color='C3', lw=1.0)
            ax.plot(wave, cont[i,:], zorder=5, color='C9', lw=1.0)

            ax.text(0.5, -0.2, r'$\lambda_{\rm obs}\ [{\rm \AA}]$',
                    ha='center', va='center', transform=ax.transAxes)




            lambda_limits = halpha_limits*(1+select_hdu['PAR'].data['Z'][i])
            indx = numpy.logical_not(numpy.ma.getmaskarray(flux)[i,:]) \
                        & (wave > lambda_limits[0]) & (wave < lambda_limits[1])
            mod_lim = [0,1] if numpy.sum(indx) == 0 \
                        else [numpy.ma.amin(model[i,indx]), numpy.ma.amax(model[i,indx])]
            Df = (mod_lim[1]-mod_lim[0])*1.5
            flux_limits = numpy.mean(mod_lim) + numpy.array([-Df/2, Df/2])

            ax = fig.add_axes([0.54, 0.08, 0.19, 0.27])
            ax.minorticks_on()
            ax.tick_params(which='major', length=10, direction='in', top=True, right=True)
            ax.tick_params(which='minor', length=5, direction='in', top=True, right=True)
            ax.set_xlim(lambda_limits)
            ax.set_ylim(flux_limits)
            ax.grid(True, which='major', color='0.8', zorder=0, linestyle='-')
            ax.plot(wave, flux[i,:], zorder=2, color='k', lw=0.5)
            ax.plot(wave, stellar[i,:], zorder=3, color='C1', lw=1.0)
            ax.plot(wave, model[i,:], zorder=4, color='C3', lw=1.0)
            ax.plot(wave, cont[i,:], zorder=5, color='C9', lw=1.0)

            ax.text(0.5, -0.2, r'$\lambda_{\rm obs}\ [{\rm \AA}]$',
                    ha='center', va='center', transform=ax.transAxes)




            lambda_limits = ca_limits*(1+select_hdu['PAR'].data['Z'][i])
            indx = numpy.logical_not(numpy.ma.getmaskarray(flux)[i,:]) \
                        & (wave > lambda_limits[0]) & (wave < lambda_limits[1])
            mod_lim = [0,1] if numpy.sum(indx) == 0 \
                        else [numpy.ma.amin(model[i,indx]), numpy.ma.amax(model[i,indx])]
            Df = (mod_lim[1]-mod_lim[0])*1.5
            flux_limits = numpy.mean(mod_lim) + numpy.array([-Df/2, Df/2])

            ax = fig.add_axes([0.78, 0.08, 0.19, 0.27])
            ax.minorticks_on()
            ax.tick_params(which='major', length=10, direction='in', top=True, right=True)
            ax.tick_params(which='minor', length=5, direction='in', top=True, right=True)
            ax.set_xlim(lambda_limits)
            ax.set_ylim(flux_limits)
            ax.grid(True, which='major', color='0.8', zorder=0, linestyle='-')
            ax.plot(wave, flux[i,:], zorder=2, color='k', lw=0.5)
            ax.plot(wave, stellar[i,:], zorder=3, color='C1', lw=1.0)
            ax.plot(wave, model[i,:], zorder=4, color='C3', lw=1.0)
            ax.plot(wave, cont[i,:], zorder=5, color='C9', lw=1.0)

            ax.text(0.5, -0.2, r'$\lambda_{\rm obs}\ [{\rm \AA}]$',
                    ha='center', va='center', transform=ax.transAxes)


            pdf.savefig(orientation='landscape')
            fig.clear()
            pyplot.close(fig)
            
#            j += 1
#            if j == 10:
#                break

    print('{0}/{0}'.format(nspec))

if __name__ == '__main__':
    t = time.perf_counter()
    main()
    print('Elapsed time: {0} seconds'.format(time.perf_counter() - t))






# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""
Defines a class to keep track of MaNGA MPL dependencies and versioning.

.. note::
    This class was created for the survey-level execution of the DAP.
    The dependencies will necessarily grow, and will be handled using
    module files.  So, even as it is, this class has little utility.

.. todo::
    Deprecate this. I'm not sure it's actually used (or useful)
    anymore.

----

.. include license and copyright
.. include:: ../include/copy.rst

----

.. include common links, assuming primary doc root is up one directory
.. include:: ../include/links.rst
"""
import sys
import os
import warnings
from distutils.version import StrictVersion

# For version checking
import numpy
import scipy
import matplotlib
import astropy
import pydl

# TODO: Add DAP version

class MaNGAMPL:
    """

    Create an object that keeps track of IDLUTILS, SDSS_ACCESS, MANGACORE, and
    MANGADRP dependencies for a given MPL and other intermediate analyses.
    Dependencies are selected by the "MPL" version as provided in the
    following table:

    **SDSS dependencies**
    +--------+-------------+------------+--------+--------+--------+
    |  MPL   | SDSS_ACCESS |  IDLUTILS  |   CORE |   DRP  |   DAP  |
    +========+=============+============+========+========+========+
    | MPL-1  |        None |    v5_5_16 | v1_0_0 | v1_0_0 |   None |
    +--------+-------------+------------+--------+--------+--------+
    | MPL-2  |        None |    v5_5_17 | v1_0_0 | v1_1_2 |   None |
    +--------+-------------+------------+--------+--------+--------+
    | v1_2_0 |        None |    v5_5_19 | v1_1_0 | v1_2_0 |   None |
    +--------+-------------+------------+--------+--------+--------+
    | MPL-3  |        None |    v5_5_22 | v1_1_0 | v1_3_3 | v1_0_0 |
    +--------+-------------+------------+--------+--------+--------+
    | MPL-4  |       0.0.0 |    v5_5_23 | v1_2_0 | v1_5_1 |  1.1.1 |
    +--------+-------------+------------+--------+--------+--------+
    | MPL-5  |       0.2.0 |    v5_5_26 | v1_3_1 | v2_0_1 |  2.0.2 |
    +--------+-------------+------------+--------+--------+--------+
    | MPL-6  |        None |    v5_5_30 | v1_6_0 | v2_3_1 |  2.1.3 |
    +--------+-------------+------------+--------+--------+--------+
    | MPL-7  |        None |    v5_5_32 | v1_6_2 | v2_4_3 |  2.2.1 |
    +--------+-------------+------------+--------+--------+--------+

    **Python versions**
    +--------+--------+--------+--------+------------+---------+-------+----------------+
    | MPL    | python | numpy  |  scipy | matplotlib | astropy |  pydl | Fwd compatible |
    +========+========+========+========+============+=========+=======+================+
    | MPL-5  |  3.5.1 | 1.11.0 | 0.17.1 |       None |   1.1.2 | 0.5.0 |            yes |
    +--------+--------+--------+--------+------------+---------+-------+----------------+
    | MPL-6  |  3.6.3 | 1.13.3 |  1.0.0 |       None |   2.0.2 | 0.6.0 |            N/A |
    +--------+--------+--------+--------+------------+---------+-------+----------------+
    | MPL-7  |  3.6.3 | 1.13.3 |  1.0.0 |       None |   2.0.2 | 0.6.0 |            N/A |
    +--------+--------+--------+--------+------------+---------+-------+----------------+

    .. note::
        - "Fwd compatible" means that you can run this version of the
          DAP with the most recent version without needed to revert the
          python packages to their previous versions.
        - Only the survey-level routines are dependent on MANGACORE; the
          core DAP processing software only depends on the listed python
          packages.

    .. todo::
        - Change python dependencies to "minimum" versions?

    .. warning::
        - SDSS_ACCESS dependencies are only because of the plotting
          package that has since moved into Marvin.
        - For MPL-1, the MANGACORE version was set to trunk.  Here it
          defaults to v1_0_0.
        - Explicit dependence on matplotlib was introduced in MPL-6.

    Args:
        version (str): (**Optional**) Version as selected via matching
            to the "MPL" string in the table listed above. Default:
            MPL-6.
        strictver (bool): (**Optional**) Strictly check the version
            requirements for this version of the dap.  Default is True,
            meaning the code will raise an exception if the expected
            version are not the same as the environment.

    Raises:
        ValueError: Raised if the provided *version* is not one of the
            MPL values in the above table.

    Attributes:
        mplver (str): Version as selected via matching to the "MPL"
            string in the table listed above. Default: MPL-5.
        accessver (str): SDSS_ACCESS version.
        idlver (str): IDLUTILS version.
        corever (str): MANGACORE version.
        drpver (str): MANGADRP version.
        dapver (str): MANGADAP version.
        pythonver (str): Python version
        numpyver (str): Numpy version
        scipyver (str): Scipy version
        matplotlibver (str): Matplotlib version
        astropyver (str): Astropy version
        pydlver (str): Pydl version
        strictver (bool): Strictly check the version requirements for
            this version of the dap.  If True, an exception is raised if
            the expected version are not the same as the environment.

    """
    def __init__(self, version=None, strictver=True):
        # Default to current environment
        self.mplver = 'env' if version is None else version

        mpls = self._available_mpls()
        mpli = numpy.where(mpls[:,0] == self.mplver)
        if len(mpli[0]) == 0:
            mpls = self._available_mpls(write=True)
            raise ValueError('{0} is not an available MPL!'.format(self.mplver))
           
        n = mpls.shape[0] 
        i = 0
        self.mplver = mpls[mpli,i].ravel()[0]
        i += 1
        self.accessver = mpls[mpli,i].ravel()[0]
        i += 1
        self.idlver = mpls[mpli,i].ravel()[0]
        i += 1
        self.corever = mpls[mpli,i].ravel()[0]
        i += 1
        self.drpver = mpls[mpli,i].ravel()[0]
        i += 1
        self.dapver = mpls[mpli,i].ravel()[0]
        i += 1
        self.pythonver = mpls[mpli,i].ravel()[0]
        i += 1
        self.numpyver = mpls[mpli,i].ravel()[0]
        i += 1
        self.scipyver = mpls[mpli,i].ravel()[0]
        i += 1
        self.matplotlibver = mpls[mpli,i].ravel()[0]
        i += 1
        self.astropyver = mpls[mpli,i].ravel()[0]
        i += 1
        self.pydlver = mpls[mpli,i].ravel()[0]
        i += 1

        # Treat versioning strictly
        self.strictver = strictver


    def _available_mpls(self, write=False):
        """
        Return a list of the available MPLs to analyze, and printing it if requested.

        .. todo:
            Have this read a config file!

        Args:
            write (bool): Write the list of available MPLs to the
            stdout.

        Returns:
            numpy.array: A string array with the MPL dependencies
        """
        #                             MPL SDSS_ACCESS  IDLUTILS      CORE       DRP       DAP
        mpl_def = numpy.array([ [ 'MPL-1',      None, 'v5_5_16', 'v1_0_0', 'v1_0_0',     None,
        #                            python     numpy     scipy  matplotlib  astropy      pydl
                                       None,     None,     None,       None,    None,     None ],
                                [ 'MPL-2',      None, 'v5_5_17', 'v1_0_0', 'v1_1_2',     None,
                                       None,     None,     None,       None,    None,     None ],
                                ['v1_2_0',      None, 'v5_5_19', 'v1_1_0', 'v1_2_0',     None,
                                       None,     None,     None,       None,    None,     None ],
                                [ 'MPL-3',      None, 'v5_5_22', 'v1_1_0', 'v1_3_3', 'v1_0_0',
                                       None,     None,     None,       None,    None,     None ],
                                [ 'MPL-4',   '0.0.0', 'v5_5_23', 'v1_2_0', 'v1_5_1',  '1.1.1',
                                       None,     None,     None,       None,    None,     None ],
                                [ 'MPL-5',   '0.2.0', 'v5_5_26', 'v1_3_1', 'v2_0_1',  '2.0.2',
                                    '3.5.1', '1.11.0', '0.17.1',       None,  '1.1.2', '0.5.0' ],
                                [ 'MPL-6',      None, 'v5_5_30', 'v1_6_0', 'v2_3_1',  '2.1.3',
                                    '3.6.3', '1.13.3',    '1.0',       None,  '2.0.2', '0.6.0' ],
                                [ 'MPL-7',      None, 'v5_5_32', 'v1_6_2', 'v2_4_3',  '2.2.1',
                                    '3.6.3', '1.13.3',    '1.0',       None,  '2.0.2', '0.6.0' ],
                                ['env'] + self.get_environment_versions()
                                ])

        nmpl = mpl_def.shape[1]
        if write:
            print('Available MPLs:')
            print('{0:>7} {1:>7} {2:>8} {3:>6} {4:>6} {5:>6} {6:>6} {7:>6} {8:>6} {9:>10}'
                  ' {10:>7} {11:>6}'.format('MPL', 'SDSSACC', 'IDLUTILS', 'CORE', 'DRP', 'DAP',
                                            'PYTHON', 'NUMPY', 'SCIPY', 'MATPLOTLIB', 'ASTROPY',
                                            'PYDL'))
            for x in mpl_def[0:nmpl,:]:
                _x = x.astype(str)
                print('{0:>7} {1:>7} {2:>8} {3:>6} {4:>6} {5:>6} {6:>6} {7:>6} {8:>6} {9:>10}'
                      ' {10:>7} {11:>6}'.format(*_x))

        return mpl_def


    def _version_mismatch(self, package, required_ver, environment_ver, quiet=False):
        """
        A version mismatch was detected.  Handle based on
        :attr:`strictver`.
        """
        if self.strictver:
            raise EnvironmentError('Environment mismatch for {0}: expected {1}, found {2}'.format(
                                   package, required_ver, environment_ver))
        elif not quiet:
            warnings.warn('Environment mismatch for {0}: expected {1}, found {2}'.format(package,
                          required_ver, environment_ver))


    def get_environment_versions(self):
        # These will throw KeyErrors if the appropriate environmental variables do not exist
        try:
            accessver_env = os.environ['SDSS_ACCESS_DIR'].split('/')[-1]
        except KeyError:
            accessver_env = None
        try:
            idlver_env = os.environ['IDLUTILS_DIR'].split('/')[-1]
        except KeyError:
            idlver_env = None
        python_ver = '.'.join([ str(v) for v in sys.version_info[:3]])

        return [ accessver_env, idlver_env, os.getenv('MANGACORE_VER'),
                 os.getenv('MANGADRP_VER'), os.getenv('MANGADAP_VER'),
                python_ver, numpy.__version__, scipy.__version__, matplotlib.__version__,
                astropy.__version__, pydl.__version__]


    def verify_versions(self, quiet=True):
        """
        Verify that the defined MPL versions are consistent with the
        system where the DAP is being run.
        """
        env = self.get_environment_versions()

        if not quiet:
            print('Current environment: ')
#            print('    SDSS_ACCESS: {0}'.format(env[0]))
            print('       IDLUTILS: {0}'.format(env[1]))
            print('      MANGACORE: {0}'.format(env[2]))
            print('       MANGADRP: {0}'.format(env[3]))
            print('       MANGADAP: {0}'.format(env[4]))
            print('         PYTHON: {0}'.format(env[5]))
            print('          NUMPY: {0}'.format(env[6]))
            print('          SCIPY: {0}'.format(env[7]))
            print('     MATPLOTLIB: {0}'.format(env[8]))
            print('        ASTROPY: {0}'.format(env[9]))
            print('           PYDL: {0}'.format(env[10]))

        # Check versions
#        if self.accessver != env[0]:
#            self._version_mismatch('SDSS_ACCESS', self.accessver, env[0], quiet=quiet)

        if self.idlver != env[1]:
            self._version_mismatch('IDLUTILS', self.idlver, env[1], quiet=quiet)

        if self.corever != env[2]:
            self._version_mismatch('MANGACORE', self.corever, env[2], quiet=quiet)

        if self.drpver != env[3]:
            self._version_mismatch('MANGADRP', self.drpver, env[3], quiet=quiet)

#        if self.dapver != env[4]:
#            self._version_mismatch('MANGADAP', self.dapver, env[4], quiet=quiet)

        if self.pythonver is not None and StrictVersion(env[5]) != StrictVersion(self.pythonver):
            self._version_mismatch('python', self.pythonver, env[5], quiet=quiet)

        if self.numpyver is not None and StrictVersion(env[6]) != StrictVersion(self.numpyver):
            self._version_mismatch('numpy', self.numpyver, env[6], quiet=quiet)

        if self.scipyver is not None and StrictVersion(env[7]) != StrictVersion(self.scipyver):
            self._version_mismatch('scipy', self.scipyver, env[7], quiet=quiet)

#        if self.matplotlibver is not None \
#                and StrictVersion(env[8]) != StrictVersion(self.matplotlibver):
#            self._version_mismatch('matplotlib', self.matplotlibver, env[8], quiet=quiet)

        if self.astropyver is not None and StrictVersion(env[9]) != StrictVersion(self.astropyver):
            self._version_mismatch('astropy', self.astropyver, env[9], quiet=quiet)

        if self.pydlver is not None and StrictVersion(env[10]) != StrictVersion(self.pydlver):
            self._version_mismatch('pydl', self.pydlver, env[10], quiet=quiet)


    def show(self):
        """Print the available MPL versions to stdout."""

        print('{0}: IDLUTILS:{1}; COREVER:{2}; DRPVER:{3}; DAPVER:{4}'.format(
              self.mplver, self.idlver, self.corever, self.drpver, self.dapver))


